#!/usr/bin/perl

# params: filename [time_limit]

use strict;

$| = 1;

my $filename=shift(@ARGV);

my $time_limit=300;
if (@ARGV>0) { $time_limit=shift(@ARGV); };

my @prolog_params=();
my $err="";


####################### FIND TYPES IN THE PROBLEM

my $t_int=0;
my $t_rat=0;
my $t_real=0;

if (!open(IN,"<$filename")) { $err="Cannot read $filename"; output(); }
while(<IN>) { 
  my $input=$_;
  if ($input =~ /(.*)%/) { $input=$1; };
  if ($input =~ /\$int/) { $t_int=1; };
  if ($input =~ /\$rat/) { $t_rat=1; };
  if ($input =~ /\$real/) { $t_real=1; };
 } 
close(IN);

my $t_pure="";  # "pure type" = all symbols and variables of the same type (else eq "")
if ($t_int==1 && $t_rat==0 && $t_real==0) { $t_pure="Integers"; };
if ($t_int==0 && $t_rat==1 && $t_real==0) { $t_pure="Rationals"; };
if ($t_int==0 && $t_rat==0 && $t_real==1) { $t_pure="Reals"; };


####################### READ & PARSE

open(IN,"<$filename");
if ($filename =~ /\/([^\/]*)$/) { $filename=$1; };

my @types=();
my @axiomsG=();
my @axiomsP=();
my $conjectureG="";
my $conjectureP="";
my $orig_status="";
my $orig_rating="";

while (1) {

# read formula
  my $input="";
  do {  
    if (!($input.=<IN>)) { goto reading_done; };
    $input =~ s/\s//g;
	if ($input =~ /%Status:(.*)/) { $orig_status=$1; };
	if ($input =~ /%Rating:(\d\.\d\d)/) { $orig_rating=$1; };
	if ($input =~ /(.*)%/) { $input=$1; };
   } until ($input =~ /\)\./);

# parse formula
  my $x=prolog_parse($input);
  if ($x eq "include") { $err="Include not supported, use include.pl to transform the input."; output(); };
  if ($x ne "tff") { $err="$x formulas not supported (TFF only), at line $input"; output(); };
  $x=$prolog_params[1]; my $y=$prolog_params[2];
  if ($x eq "type") { $types[@types]=$y; }
    elsif ($x eq "axiom") { $axiomsG[@axiomsG]=tfa2math($y,"G"); $axiomsP[@axiomsP]=tfa2math($y,"P");}
    elsif ($x eq "conjecture") { $conjectureG=tfa2math($y,"G"); $conjectureP=tfa2math($y,"P");}
    else { $err="Unknown formula type $x"; };	
}


####################### MATHEMATICA INPUT

reading_done:
# axioms + conjecture -> single formula

my $flaG;  # formula with guards
my $flaP;  # formula without guards (non-trivial only if pure type)
if (@axiomsG>0) {
	my $xG="";my $xP="";
	my $i;for($i=0;$i<@axiomsG-1;$i++) { $xG.="(".$axiomsG[$i].")&&"; $xP.="(".$axiomsP[$i].")&&";};
	$flaG="Implies[$xG($axiomsG[$i]),$conjectureG]";
	$flaP="Implies[$xP($axiomsP[$i]),$conjectureP]";
 }
 else { $flaG=$conjectureG; $flaP=$conjectureP; };

# Mathematica code

my @lines=();
my @limits;

$time_limit*=0.95;
if ($time_limit<=16) { 
	for(my $i=0;$i<18; $i++) { $limits[$i]=$time_limit*0.062; }; 
	$limits[18]=0.1; $limits[19]=0.1; 
 } else { 
	for(my $i=0;$i<16; $i++) { $limits[$i]=1; };
	for(my $i=16;$i<20; $i++) { $limits[$i]=($time_limit-15)*0.33; };
 };
 
$lines[@lines]="h2wo4=Reduce[$flaG]";
$lines[@lines]="FullSimplify[h2wo4]";
$lines[@lines]="Reduce[!h2wo4]";
$lines[@lines]="h2wo4=Reduce[!($flaG)]";
$lines[@lines]="FullSimplify[h2wo4]";
$lines[@lines]="Reduce[!h2wo4]";

$lines[@lines]="h2wo4=FullSimplify[$flaG]";
$lines[@lines]="Reduce[h2wo4]";
$lines[@lines]="If[h2wo4[[0]]===Exists,FindInstance[h2wo4[[2]],h2wo4[[1]]],Print[\"VOID\"]]";
$lines[@lines]="h2wo4=FullSimplify[!($flaG)]";
$lines[@lines]="Reduce[h2wo4]";
$lines[@lines]="If[h2wo4[[0]]===Exists,FindInstance[h2wo4[[2]],h2wo4[[1]]],Print[\"VOID\"]]";

if ($t_pure eq "") {
	$lines[@lines]="Print[\"VOID\"]";
	$lines[@lines]="Print[\"VOID\"]";
	$lines[@lines]="Print[\"VOID\"]";
	$lines[@lines]="Print[\"VOID\"]";
	$lines[@lines]="Print[\"VOID\"]";
	$lines[@lines]="Print[\"VOID\"]";
 }
 else {
	$lines[@lines]="Reduce[$flaP,$t_pure]";
	$lines[@lines]="If[%===False,Print[\"VOID\"],Reduce[!%,$t_pure]]";
	$lines[@lines]="Reduce[!($flaP),$t_pure]";
	$lines[@lines]="If[%===False,Print[\"VOID\"],Reduce[!%,$t_pure]]";

	$lines[@lines]="h2wo4=$flaP;If[h2wo4[[0]]===Exists,FindInstance[h2wo4[[2]],h2wo4[[1]],$t_pure],Print[\"VOID\"]]";
	$lines[@lines]="h2wo4=!h2wo4;If[h2wo4[[0]]===Exists,FindInstance[h2wo4[[2]],h2wo4[[1]],$t_pure],Print[\"VOID\"]]";
 }
 
$lines[@lines]="Reduce[$flaG]";
$lines[@lines]="Reduce[!($flaG)]";

# goto output
output();


####################### PRINT AND QUIT

sub output {

 print "Print[\"$filename\"]\n";
 print "Print[\"$orig_status\"]\n";
 print "Print[\"$orig_rating\"]\n";
 print "Print[\"$t_pure\"]\n";
 print "Print[\"@types\"]\n";
 
 if ($err eq "") {
    print "Print[\"Parsing OK.\"]\n";
    print "$flaG//InputForm\n";
    if ($t_pure ne "") { print "$flaP//InputForm\n"; } else { print "Print[\"VOID\"]\n"; };
    for(my $i=0;$i<@lines;$i++) { 
		print "TimeConstrained[$lines[$i],$limits[$i]]//InputForm\n";
	 }
   }
   else {
    print "Print[\"ERROR $err\"]\n";
   }
 
 close(IN);
 exit;
}


######################## PARSING

# TFA formula -> Mathematica expression
sub tfa2math {
 my $f=depar(shift(@_)); 
 my $purity=shift(@_);
 
 if ($t_pure eq "" && $purity eq "P") { return ""; };
 
#print "$f\n"; ###

 # forall,exists
 if (substr($f,0,1) eq "!" || substr($f,0,1) eq "?") { 
	my $q=prolog_parse(substr($f,0,length($f)-1));
	$f =~ /.\[[^\]]*\]:(.*)/;
    $f="($1)";
	my $v="";
	my $x; foreach $x (@prolog_params) { 
	  $x=~/(.*):(.*)/;
	  if ($purity eq "G") { if ($q eq "?") { $f.="&($2($1))"; } else { $f.="|(~$2($1))"; }; };
	  $v.=symbol2math($1).",";
	 }; 
	$v=substr($v,0,length($v)-1);
	$f=tfa2math($f,$purity);
	if ($q eq "?") { return "Exists[{$v},$f]" } else { return "ForAll[{$v},$f]" };
  };
 
 # or,and,=>,<=>
 my $p=0;my $i=-1;
 while ($i<length($f) && $p>=0) {
	$i++;
	if (substr($f,$i,1) eq "(") { $p++; };
	if (substr($f,$i,1) eq ")") { $p--; };
	if ($p==0 && (substr($f,$i,1) eq "&" || substr($f,$i,1) eq "|" || substr($f,$i,2) eq "=>" || substr($f,$i,3) eq "<=>")) { $p=-1; };
  };
 if ($p==-1 && substr($f,$i,1) eq "&") { my $s=tfa2math(substr($f,0,$i),$purity); my $t=tfa2math(substr($f,$i+1,length($f)-$i-1),$purity); return "($s) && ($t)"; };
 if ($p==-1 && substr($f,$i,1) eq "|") { my $s=tfa2math(substr($f,0,$i),$purity); my $t=tfa2math(substr($f,$i+1,length($f)-$i-1),$purity); return "($s) || ($t)"; };
 if ($p==-1 && substr($f,$i,1) eq "=") { my $s=tfa2math(substr($f,0,$i),$purity); my $t=tfa2math(substr($f,$i+2,length($f)-$i-2),$purity); return "Implies[$s,$t]"; };
 if ($p==-1 && substr($f,$i,1) eq "<") { my $s=tfa2math(substr($f,0,$i),$purity); my $t=tfa2math(substr($f,$i+3,length($f)-$i-3),$purity); return "Equivalent[$s,$t]"; };

 # neg
 if (substr($f,0,1) eq "~") { my $s=tfa2math(substr($f,1,length($f)-1),$purity);return "Not[$s]"; };

#else: atomic or term 

 # =,!=
 if ($f =~ /(.*)!=(.*)/) { my $s=tfa2math($1,$purity);my $t=tfa2math($2,$purity);return "$s != $t"; };
 if ($f =~ /(.*)=(.*)/) { my $s=tfa2math($1,$purity);my $t=tfa2math($2,$purity);return "$s == $t"; };

#else: term or predicate
 my $h=prolog_parse($f);
 my @params=@prolog_params;
 
 # built-in symbols
 if (substr($h,0,1) eq "\$") {
  if ($h eq "\$int") 
   { my $s=tfa2math($params[0],$purity);return "Element[$s,Integers]" };
  if ($h eq "\$rat") 
   { my $s=tfa2math($params[0],$purity);return "Element[$s,Rationals]" };
  if ($h eq "\$real") 
   { my $s=tfa2math($params[0],$purity);return "Element[$s,Reals]" };
  if ($h eq "\$is_int") 
   { my $s=tfa2math($params[0],$purity);return "Element[$s,Integers]" };
  if ($h eq "\$is_rat") 
   { my $s=tfa2math($params[0],$purity);return "Element[$s,Rationals]" };
  if ($h eq "\$is_real") 
   { my $s=tfa2math($params[0],$purity);return "Element[$s,Reals]" };
  if ($h eq "\$to_int") 
   { my $s=tfa2math($params[0],$purity);return "Floor[$s]" };
  if ($h eq "\$to_rat") 
   { my $s=tfa2math($params[0],$purity);return "$s" };
  if ($h eq "\$to_real") 
   { my $s=tfa2math($params[0],$purity);return "$s" };

  if ($h eq "\$less") 
   { my $s=$params[0];my $t=$params[1];$s=tfa2math($s,$purity);$t=tfa2math($t,$purity);return "$s < $t" };
  if ($h eq "\$lesseq") 
   { my $s=$params[0];my $t=$params[1];$s=tfa2math($s,$purity);$t=tfa2math($t,$purity);return "$s <= $t" };
  if ($h eq "\$greater") 
   { my $s=$params[0];my $t=$params[1];$s=tfa2math($s,$purity);$t=tfa2math($t,$purity);return "$s > $t" };
  if ($h eq "\$greatereq") 
   { my $s=$params[0];my $t=$params[1];$s=tfa2math($s,$purity);$t=tfa2math($t,$purity);return "$s >= $t" };
  if ($h eq "\$uminus") 
   { my $s=tfa2math($params[0],$purity);return "-($s)" };
  if ($h eq "\$sum") 
   { my $s=$params[0];my $t=$params[1];$s=tfa2math($s,$purity);$t=tfa2math($t,$purity);return "($s) + ($t)" };
  if ($h eq "\$difference") 
   { my $s=$params[0];my $t=$params[1];$s=tfa2math($s,$purity);$t=tfa2math($t,$purity);return "($s) - ($t)" };
  if ($h eq "\$product") 
   { my $s=$params[0];my $t=$params[1];$s=tfa2math($s,$purity);$t=tfa2math($t,$purity);return "($s) * ($t)" };
  if ($h eq "\$evaleq") 
   { my $s=$params[0];my $t=$params[1];$s=tfa2math($s,$purity);$t=tfa2math($t,$purity);return "$s == $t" };

  if ($h eq "\$true") { return "True" };
  if ($h eq "\$false") { return "False" };
  $err="Symbol $h not supported."; output();
  }
 
 # free symbols
 $h=symbol2math($h);
 if (@params==0) { return $h; };
 $h=$h."[";
 for($i=0;$i<@params-1;$i++) { $h.=tfa2math($params[$i],$purity).","; };
 return $h.tfa2math($params[$i])."]";
}


# identifier, number -> Mathematica expression
sub symbol2math {
 my $t=shift(@_); 

 # number
 if ($t =~ /^-\d/ || $t =~ /^\d/) {  
		if ($t=~/(.*)e(.*)/ || $t=~/(.*)E(.*)/) { my $x=symbol2math($1); $t="$x*10^($2)"; };
		if ($t=~/(.*)\.(.*)/) { my $x=length($2); $t="$1$2/10^$x"; };
	}
  # variable or free symbol
  else { 
		my $p=0; my $x;
		foreach $x (@types) { if ($x =~ /$t:([^>]*)>\$o/) { $p=1; } };
		$t = "h2wo4".$t;
		$t =~ s/_/UNDErsCORE/g;
		if ($p==1) { $t="True==".$t; };
	}
 return $t;
}


################## PROLOG TERMS

# vrati vsecko pred prvni zavorkou
sub prolog_head {
 my $s=shift(@_);
 my $i=0;
 
 while (substr($s,$i,1) ne "(" && $i<length($s)) { $i++; };
 if (substr($s,$i,1) eq "(") { return substr($s,0,$i); };
 return $s;
}


# vrati hlavicku a naplni @prolog_params parametry, funguje i pro list
sub prolog_parse {
 my $s=shift(@_);
 my $i=0;
 my $z=0;
 @prolog_params=();

 while (substr($s,$i,1) ne "(" && substr($s,$i,1) ne "[" && substr($s,$i,1) ne "," && $i<length($s)) { $i++; };
 if (substr($s,$i,1) eq ",") { return substr($s,0,$i); };
 if (substr($s,$i,1) ne "(" && substr($s,$i,1) ne "[") { return $s; };
 my $odpoved=substr($s,0,$i);
 my $j=0;
 $z=1;
 my $zi=$i+1;
 my $apo=0;
 while ($z>0 && $i<length($s))
  {
   $i++;
   if (substr($s,$i,1) eq "'") { $apo=1-$apo; };
   if (substr($s,$i,1) eq "[") { $z++; };
   if (substr($s,$i,1) eq "]") { $z--; };   
   if (substr($s,$i,1) eq "(") { $z++; };
   if (substr($s,$i,1) eq ")") { $z--; };   
   if (substr($s,$i,1) eq "," && $z==1 && $apo==0) { $prolog_params[$j]=substr($s,$zi,$i-$zi); $j++; $zi=$i+1; };   
  }   
 
 if ($i!=$zi) { $prolog_params[$j]=substr($s,$zi,$i-$zi); };
 if ($z==0) { return $odpoved; } else { $err="Fatal error: $s doesn't contain a prolog term."; };
}

sub depar {
 my $s=shift(@_);
 my $t;
 do { $t=$s;$s=depar1($t); } until ($t eq $s);
 return $s;
} 

sub depar1 {
 my $s=shift(@_);

 if (!($s =~ /^\(.*\)$/)) { return $s; };
 my $i=1;
 my $z=1;
 while ($z>0 && $i<length($s))
  {
   if (substr($s,$i,1) eq "(") { $z++; };
   if (substr($s,$i,1) eq ")") { $z--; };   
   $i++;
  }
 if ($i==length($s)) { return substr($s,1,length($s)-2); } else { return $s; };
}
